import os
from buzhug import Base
import datetime
from threading import Thread
import urllib
import simplejson

class ScoreManager(object):
	def __init__(self, db_path):
		self.db_path = db_path
		self.users = Base(os.path.join(db_path, 'users')).create(('user_id', str), mode="open")
		# attempt to get a single entry from the base
		try:
			self.user_id = self.users[0].user_id
		except IndexError:
			from uuid import uuid4
			from hashlib import sha1
			from base64 import b32encode
			self.user_id = b32encode(sha1(uuid4().bytes).digest()).lower()
			self.users.insert(user_id=self.user_id)
		self.games = Base(os.path.join(db_path, 'games')).create(
			('user_id', str),
			('mode', str),
			('score', int),
			('depth', float),
			('created_at', datetime.datetime),
			mode="open")
		self.stats = Base(os.path.join(db_path, 'stats')).create(
			('user_id', str),
			('mode', str),
			('words_formed', int),
			('letters_used', int),
			('total_depth', float),
			('seconds_elapsed', int),
			('games_played', int),
			mode="open")
		self.words = Base(os.path.join(db_path, 'words')).create(
			('user_id', str),
			('word', str),
			('mode', str),
			('score', int),
			('created_at', datetime.datetime),
			mode="open")
		self.settings = Base(os.path.join(db_path, 'settings')).create(
			('user_id', str),
			('tutorial', bool),
			('music', bool),
			('sfx', bool),
			('use_remote', bool),
			('first_time', bool),
			('show_version_warning', bool),
			mode="open")

		# patch in new fields
		try:
			x = self.settings[0].show_version_warning
		except AttributeError:
			self.settings.add_field('show_version_warning', bool)
		except IndexError:
			pass

		# create default settings if necessary
		try:
			settings = self.settings[0]
		except IndexError:
			self.settings.insert(user_id=self.user_id, tutorial=True, music=True,
					sfx=True, use_remote=True, first_time=True,
					show_version_warning=False)

		self.quests = Base(os.path.join(db_path, 'quests')).create(
			('user_id', str),
			('name', str),
			('level', int),
			('progress', int),
			('complete', bool),
			('data', str),
			mode="open")

	def get_user_id(self):
		return self.user_id
	
	def get_setting(self, setting):
		prefs = self.settings(user_id=self.user_id)[0]
		return getattr(prefs, setting)

	def set_setting(self, setting, val):
		prefs = self.settings(user_id=self.user_id)[0]
		prefs.update(**{setting: val})

	def record_game(self, mode, score, depth):
		self.games.insert(user_id=self.user_id, mode=mode, score=score,
				depth=depth, created_at=datetime.datetime.utcnow())

	def record_words(self, mode, words):
		for word_d in words:
			if len(self.words(word=word_d['word'], mode=mode)) == 0:
				self.words.insert(user_id=self.user_id, mode=mode,
						score=word_d['score'], word=word_d['word'],
						created_at=datetime.datetime.utcnow())

	def update_stats(self, mode, words_formed, letters_used, depth,
			seconds_elapsed):
		try:
			stats = self.stats(user_id=self.user_id, mode=mode)[0]
		except IndexError:
			self.stats.insert(user_id=self.user_id, mode=mode,
					words_formed=words_formed, letters_used=letters_used,
					total_depth=depth, seconds_elapsed=seconds_elapsed,
					games_played=1)
			return
		stats.update(
			words_formed=stats.words_formed+words_formed,
			letters_used=stats.letters_used+letters_used,
			total_depth=stats.total_depth+depth,
			seconds_elapsed=stats.seconds_elapsed+seconds_elapsed,
			games_played=stats.games_played+1)

	def get_high_scores(self, mode=None, count=5):
		if mode is None:
			rset = self.games(user_id=self.user_id).sort_by("-score")[:count]
		else:
			rset = self.games(user_id=self.user_id, mode=mode).sort_by("-score")[:count]
		return rset

	def get_best_depths(self, mode=None, count=5):
		if mode is None:
			rset = self.games(user_id=self.user_id).sort_by("-depth")[:count]
		else:
			rset = self.games(user_id=self.user_id, mode=mode).sort_by("-depth")[:count]
		return rset

	def get_best_words(self, mode=None, count=5):
		if mode is None:
			rset = self.words(user_id=self.user_id).sort_by('-score')[:count]
		else:
			rset = self.words(user_id=self.user_id, mode=mode).sort_by('-score')[:count]
		return rset

	def get_stats(self, mode=None):
		rec = {'words_formed': 0, 'letters_used': 0, 'total_depth': 0,
				'seconds_elapsed': 0, 'games_played': 0}
		if mode is None:
			rset = self.stats(user_id=self.user_id)
		else:
			rset = self.stats(user_id=self.user_id, mode=mode)
		for row in rset:
			rec['words_formed'] += row.words_formed
			rec['letters_used'] += row.letters_used
			rec['total_depth'] += row.total_depth
			rec['seconds_elapsed'] += row.seconds_elapsed
			rec['games_played'] += row.games_played
		return rec

	def save(self):
		pass

def _record_game_call(base_url, user_id, mode, score, depth, callback):
	resp = urllib.urlopen(base_url + "/games", data=urllib.urlencode({
		'user_id': user_id,
		'mode': mode,
		'score': score,
		'depth': depth
	}))
	content = simplejson.loads(resp.read())
	if 'error' in content:
		callback(success=False, leaderboard={})
	else:
		callback(success=True, leaderboard=content['leaderboard'])

def _record_words_call(base_url, user_id, mode, words, callback):
	success = True
	resp = urllib.urlopen(base_url + "/words", data=urllib.urlencode({
		'user_id': user_id,
		'words': simplejson.dumps(words),
		'mode': mode,
	}))
	content = simplejson.loads(resp.read())
	if 'error' in content:
		success = False
	if success:
		callback(success=True, uniques=content['uniques'])
	else:
		callback(success=False, uniques=[])

def _fetch_scores(base_url, mode, callback):
	resp = urllib.urlopen(base_url + "/games?mode=%s&sort=score" % mode)
	content = simplejson.loads(resp.read())
	high_score = content['records'][0]['score']
	resp = urllib.urlopen(base_url + "/games?mode=%s&sort=depth" % mode)
	content = simplejson.loads(resp.read())
	best_depth = content['records'][0]['depth']
	resp = urllib.urlopen(base_url + "/words?mode=%s" % mode)
	content = simplejson.loads(resp.read())
	best_word = content['records'][0]['word']
	best_word_score = content['records'][0]['score']
	callback(high_score, best_depth, best_word, best_word_score)

def _get_version(base_url, osname, client_version, callback):
	resp = urllib.urlopen(base_url + "/meta/version?version=%d&os=%s" % (
			client_version, osname))
	content = simplejson.loads(resp.read())
	callback(content['version'])

class AsyncRemoteScoreManager(object):
	def __init__(self, base_url):
		self.base_url = base_url
	def record_game(self, user_id, mode, score, depth, callback):
		thread = Thread(target=_record_game_call, args=(self.base_url, user_id,
				mode, score, depth, callback))	
		thread.start()
	def record_words(self, user_id, mode, words, callback):
		thread = Thread(target=_record_words_call, args=(self.base_url, user_id,
				mode, words, callback))
		thread.start()
	def get_high_scores(self, mode, callback):
		thread = Thread(target=_fetch_scores, args=(self.base_url, mode,
				callback))
		thread.start()
	def get_version(self, osname, client_version, callback):
		thread = Thread(target=_get_version, args=(self.base_url, osname,
				client_version, callback))
		thread.start()

